import collections
import six

from autotranslate.compat import goslate, googleapiclient

from django.conf import settings


class BaseTranslatorService:
    """
    Defines the base methods that should be implemented
    """

    def translate_string(self, text, target_language, source_language="en"):
        """
        Returns a single translated string literal for the target language.
        """
        raise NotImplementedError(".translate_string() must be overridden.")

    def translate_strings(
        self, strings, target_language, source_language="en", optimized=True
    ):
        """
        Returns a iterator containing translated strings for the target language
        in the same order as in the strings.
        :return:    if `optimized` is True returns a generator else an array
        """
        raise NotImplementedError(".translate_strings() must be overridden.")


class GoSlateTranslatorService(BaseTranslatorService):
    """
    Uses the free web-based API for translating.
    https://bitbucket.org/zhuoqiang/goslate
    """

    def __init__(self):
        assert goslate, "`GoSlateTranslatorService` requires `goslate` package"
        self.service = goslate.Goslate()

    def translate_string(self, text, target_language, source_language="en"):
        assert isinstance(text, six.string_types), "`text` should a string literal"
        return self.service.translate(text, target_language, source_language)

    def translate_strings(
        self, strings, target_language, source_language="en", optimized=True
    ):
        try:
            assert isinstance(
                strings, collections.Iterable
            ), "`strings` should a iterable containing string_types"
        except:
            assert isinstance(
                strings, collections.abc.Iterable
            ), "`strings` should a iterable containing string_types"
        translations = self.service.translate(strings, target_language, source_language)
        return translations if optimized else [_ for _ in translations]


class GoogleAPITranslatorService(BaseTranslatorService):
    """
    Uses the paid Google API for translating.
    https://github.com/google/google-api-python-client
    """

    def __init__(self, max_segments=128):
        assert (
            googleapiclient
        ), "`GoogleAPITranslatorService` requires `google-api-python-client` package"

        self.developer_key = getattr(settings, "GOOGLE_TRANSLATE_KEY", None)
        assert self.developer_key, (
            "`GOOGLE_TRANSLATE_KEY` is not configured, "
            "it is required by `GoogleAPITranslatorService`"
        )

        from googleapiclient.discovery import build

        self.service = build("translate", "v2", developerKey=self.developer_key)

        # the google translation API has a limit of max
        # 128 translations in a single request
        # and throws `Too many text segments Error`
        self.max_segments = max_segments
        self.translated_strings = []

    def translate_string(self, text, target_language, source_language="en"):
        assert isinstance(text, six.string_types), "`text` should a string literal"
        response = (
            self.service.translations()
            .list(source=source_language, target=target_language, q=[text])
            .execute()
        )
        return response.get("translations").pop(0).get("translatedText")

    def translate_strings(
        self, strings, target_language, source_language="en", optimized=True
    ):
        try:
            assert isinstance(
                strings, collections.MutableSequence
            ), "`strings` should be a sequence containing string_types"
        except:
            assert isinstance(
                strings, collections.abc.MutableSequence
            ), "`strings` should be a sequence containing string_types"
        assert (
            not optimized
        ), "optimized=True is not supported in `GoogleAPITranslatorService`"
        if len(strings) == 0:
            return []
        elif len(strings) <= self.max_segments:
            setattr(self, "translated_strings", getattr(self, "translated_strings", []))
            response = (
                self.service.translations()
                .list(source=source_language, target=target_language, q=strings)
                .execute()
            )
            self.translated_strings.extend(
                [t.get("translatedText") for t in response.get("translations")]
            )
            return self.translated_strings
        else:
            self.translate_strings(
                strings[0 : self.max_segments],
                target_language,
                source_language,
                optimized,
            )
            _translated_strings = self.translate_strings(
                strings[self.max_segments :],
                target_language,
                source_language,
                optimized,
            )

            # reset the property or it will grow with subsequent calls
            self.translated_strings = []
            return _translated_strings



"""
The `compat` module provides support for backwards compatibility with older
versions of Django/Python, and compatibility wrappers around optional packages.
"""
try:
    # Available in Python 3.1+
    import importlib
except ImportError:
    # Will be removed in Django 1.9
    from django.utils import importlib

try:
    import goslate
except ImportError:
    goslate = None
except SyntaxError:
    import sys
    import warnings
    warnings.warn('goslate disabled due lack support of Python-%s' % (
        sys.version.split()[0][:3]), RuntimeWarning)
    goslate = None

try:
    import googleapiclient
except ImportError:
    googleapiclient = None



import logging
import os
import re
from optparse import make_option

import polib
from django.conf import settings
from django.core.management.base import BaseCommand

from autotranslate.utils import get_translator

logger = logging.getLogger(__name__)

# not sure whether we actually need this
# just making this change for backward compatibility
# it was always empty anyways
# https://github.com/django/django/blob/1.9/django/core/management/base.py#L210
default_options = () if not hasattr(BaseCommand, 'option_list') \
    else BaseCommand.option_list


class Command(BaseCommand):
    help = ('autotranslate all the message files that have been generated '
            'using the `makemessages` command.')

    option_list = default_options + (
        make_option('--locale', '-l', default=[], dest='locale', action='append',
                    help='autotranslate the message files for the given locale(s) (e.g. pt_BR). '
                         'can be used multiple times.'),
        make_option('--untranslated', '-u', default=False, dest='skip_translated', action='store_true',
                    help='autotranslate the fuzzy and empty messages only.'),
        make_option('--set-fuzzy', '-f', default=False, dest='set_fuzzy', action='store_true',
                    help='set the fuzzy flag on autotranslated messages.'),
        make_option('--source-language', '-s', default='en', dest='source_language', action='store',
                    help='override the default source language (en) used for translation.'),
    )

    def add_arguments(self, parser):
        # Previously, only the standard optparse library was supported and
        # you would have to extend the command option_list variable with optparse.make_option().
        # See: https://docs.djangoproject.com/en/1.8/howto/custom-management-commands/#accepting-optional-arguments
        # In django 1.8, these custom options can be added in the add_arguments()
        parser.add_argument('--locale', '-l', default=[], dest='locale', action='append',
                            help='autotranslate the message files for the given locale(s) (e.g. pt_BR). '
                                 'can be used multiple times.')
        parser.add_argument('--untranslated', '-u', default=False, dest='skip_translated', action='store_true',
                            help='autotranslate the fuzzy and empty messages only.')
        parser.add_argument('--set-fuzzy', '-f', default=False, dest='set_fuzzy', action='store_true',
                            help='set the fuzzy flag on autotranslated messages.')
        parser.add_argument('--source-language', '-s', default='en', dest='source_language', action='store',
                            help='override the default source language (en) used for translation.')

    def set_options(self, **options):
        self.locale = options['locale']
        self.skip_translated = options['skip_translated']
        self.set_fuzzy = options['set_fuzzy']
        self.source_language = options['source_language']

    def handle(self, *args, **options):
        self.set_options(**options)

        assert getattr(settings, 'USE_I18N', False), 'i18n framework is disabled'
        assert getattr(settings, 'LOCALE_PATHS', []), 'locale paths is not configured properly'
        for directory in settings.LOCALE_PATHS:
            # walk through all the paths
            # and find all the pot files
            for root, dirs, files in os.walk(directory):
                for file in files:
                    if not file.endswith('.po'):
                        # process file only
                        # if its a pot file
                        continue

                    # get the target language from the parent folder name
                    target_language = os.path.basename(os.path.dirname(root))

                    if self.locale and target_language not in self.locale:
                        logger.info('skipping translation for locale `{}`'.format(target_language))
                        continue

                    self.translate_file(root, file, target_language)

    def translate_file(self, root, file_name, target_language):
        """
        convenience method for translating a pot file

        :param root:            the absolute path of folder where the file is present
        :param file_name:       name of the file to be translated (it should be a pot file)
        :param target_language: language in which the file needs to be translated
        """
        logger.info('filling up translations for locale `{}`'.format(target_language))

        po = polib.pofile(os.path.join(root, file_name))
        strings = self.get_strings_to_translate(po)

        # translate the strings,
        # all the translated strings are returned
        # in the same order on the same index
        # viz. [a, b] -> [trans_a, trans_b]
        tl = get_translator()
        translated_strings = tl.translate_strings(strings, target_language, self.source_language, False)
        self.update_translations(po, translated_strings)
        po.save()

    def need_translate(self, entry):
        return not entry.obsolete and (not (self.skip_translated and entry.translated()))

    def get_strings_to_translate(self, po):
        """Return list of string to translate from po file.

        :param po: POFile object to translate
        :type po: polib.POFile
        :return: list of string to translate
        :rtype: collections.Iterable[six.text_type]
        """
        strings = []
        for index, entry in enumerate(po):
            if not self.need_translate(entry):
                continue
            strings.append(humanize_placeholders(entry.msgid))
            if entry.msgid_plural:
                strings.append(humanize_placeholders(entry.msgid_plural))
        return strings

    def update_translations(self, entries, translated_strings):
        """Update translations in entries.

        The order and number of translations should match to get_strings_to_translate() result.

        :param entries: list of entries to translate
        :type entries: collections.Iterable[polib.POEntry] | polib.POFile
        :param translated_strings: list of translations
        :type translated_strings: collections.Iterable[six.text_type]
        """
        translations = iter(translated_strings)
        for entry in entries:
            if not self.need_translate(entry):
                continue

            if entry.msgid_plural:
                # fill the first plural form with the entry.msgid translation
                translation = next(translations)
                translation = fix_translation(entry.msgid, translation)
                entry.msgstr_plural[0] = translation

                # fill the rest of plural forms with the entry.msgid_plural translation
                translation = next(translations)
                translation = fix_translation(entry.msgid_plural, translation)
                for k, v in entry.msgstr_plural.items():
                    if k != 0:
                        entry.msgstr_plural[k] = translation
            else:
                translation = next(translations)
                translation = fix_translation(entry.msgid, translation)
                entry.msgstr = translation

            # Set the 'fuzzy' flag on translation
            if self.set_fuzzy and 'fuzzy' not in entry.flags:
                entry.flags.append('fuzzy')


def humanize_placeholders(msgid):
    """Convert placeholders to the (google translate) service friendly form.

    %(name)s -> __name__
    %s       -> __item__
    %d       -> __number__
    """
    return re.sub(
        r'%(?:\((\w+)\))?([sd])',
        lambda match: r'__{0}__'.format(
            match.group(1).lower() if match.group(1) else 'number' if match.group(2) == 'd' else 'item'),
        msgid)


def restore_placeholders(msgid, translation):
    """Restore placeholders in the translated message."""
    placehoders = re.findall(r'(\s*)(%(?:\(\w+\))?[sd])(\s*)', msgid)
    return re.sub(
        r'(\s*)(__[\w]+?__)(\s*)',
        lambda matches: '{0}{1}{2}'.format(placehoders[0][0], placehoders[0][1], placehoders.pop(0)[2]),
        translation)


def fix_translation(msgid, translation):
    # Google Translate removes a lot of formatting, these are the fixes:
    # - Add newline in the beginning if msgid also has that
    if msgid.startswith('\n') and not translation.startswith('\n'):
        translation = u'\n' + translation

    # - Add newline at the end if msgid also has that
    if msgid.endswith('\n') and not translation.endswith('\n'):
        translation += u'\n'

    # Remove spaces that have been placed between %(id) tags
    translation = restore_placeholders(msgid, translation)
    return translation



import pkg_resources
import sys


def _parse_version(v):
   return tuple(map(int, pkg_resources.get_distribution('django').version.split('.')))


# prevent import tests twice in dj16-py26
if sys.version_info[:2] == (2, 6) and \
   _parse_version(pkg_resources.get_distribution('django').version)[:2] == (1, 6):
   pass
else:
   from autotranslate.tests.test_translate_messages import *



import os

try:
    # python2.6
    import unittest2 as unittest
except ImportError:
    import unittest

import polib

from autotranslate.management.commands.translate_messages import humanize_placeholders, restore_placeholders, Command


class HumanizeTestCase(unittest.TestCase):
    def test_named_placeholders(self):
        self.assertEqual('foo __item__ bar', humanize_placeholders('foo %(item)s bar'))
        self.assertEqual('foo __item_name__ bar', humanize_placeholders('foo %(item_name)s bar'))

        self.assertEqual('foo % (item)s bar', humanize_placeholders('foo % (item)s bar'))

    def test_positional_placeholders(self):
        self.assertEqual('foo __item__ bar', humanize_placeholders('foo %s bar'))
        self.assertEqual('foo __number__ bar', humanize_placeholders('foo %d bar'))
        self.assertEqual('foo __item__ bar __item__', humanize_placeholders('foo %s bar %s'))
        self.assertEqual('foo __item____item__', humanize_placeholders('foo %s%s'))


class RestoreTestCase(unittest.TestCase):
    def test_restore_placeholders(self):
        self.assertEqual('baz %(item)s zilot',
                         restore_placeholders('foo %(item)s bar', 'baz __over__ zilot'))
        self.assertEqual('baz %(item_name)s zilot',
                         restore_placeholders('foo %(item_name)s bar', 'baz __item_name__ zilot'))
        self.assertEqual('baz %s zilot',
                         restore_placeholders('foo %s bar', 'baz __item__ zilot'))
        self.assertEqual('baz %s%s zilot',
                         restore_placeholders('foo %s%s bar', 'baz __item____item__ zilot'))


class POFileTestCase(unittest.TestCase):
    def setUp(self):
        cmd = Command()
        cmd.set_options(**dict(
                locale='ia',
                set_fuzzy=False,
                skip_translated=False
        ))
        self.cmd = cmd
        self.po = polib.pofile(os.path.join(os.path.dirname(__file__), 'data/django.po'))

    def test_should_read_single(self):
        strings = self.cmd.get_strings_to_translate(self.po)
        self.assertIn('Location', strings)

    def test_should_update_single(self):
        translations = ['XXXX']
        entry = self.po[0]
        self.cmd.update_translations([entry], translations)
        self.assertEqual('XXXX', entry.msgstr)
        self.assertTrue(entry.translated())

    def test_should_read_plural(self):
        strings = self.cmd.get_strings_to_translate(self.po)
        self.assertIn('City', strings)
        self.assertIn('Cities', strings)

    def test_should_update_plural(self):
        translations = ['SINGULAR', 'PLURAL']
        entry = self.po[1]
        self.cmd.update_translations([entry], translations)
        self.assertEqual('', entry.msgstr)
        self.assertEqual('SINGULAR', entry.msgstr_plural[0])
        self.assertEqual(['PLURAL'] * (len(entry.msgstr_plural) - 1),
                         [v for k, v in entry.msgstr_plural.items() if k != 0])
        self.assertTrue(entry.translated())



import os
import uuid

from distutils.core import setup
from setuptools import find_packages

try:  # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:  # for pip <= 9.0.3
    from pip.req import parse_requirements


PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))


with open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md")
) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_requirements = parse_requirements(
    os.path.join(PROJECT_ROOT, "requirements.txt"), session=uuid.uuid1()
)

# e.g. ['django', 'google-api-python-client']
requirements = [
    getattr(ir, "requirement", str(getattr(ir, "req", None)))
    for ir in install_requirements
]

setup(
    name="django-autotranslate2",
    version="1.2.2",
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    license="MIT License",
    description="A simple Django app to automatically translate the pot (`.po`) files generated by django's "
    "makemessages command using google translate.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/achillis2/django-autotranslate2",
    author="Ding Li",
    author_email="dingli@gmail.com",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import six

from autotranslate.compat import importlib
from django.conf import settings


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    Credits: https://github.com/tomchristie/django-rest-framework/blob/master/rest_framework/settings.py#L138
    """
    if val is None:
        return None
    elif isinstance(val, six.string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        parts = val.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        raise ImportError('Could not import {} for API setting {}. {}: {}.'
                          .format(val, setting_name, e.__class__.__name__, e))

def get_translator():
    """
    Returns the default translator.
    """
    TranslatorService = getattr(settings, 'AUTOTRANSLATE_TRANSLATOR_SERVICE',
                                'autotranslate.services.GoSlateTranslatorService')
    translator = perform_import(TranslatorService, 'AUTOTRANSLATE_TRANSLATOR_SERVICE')()
    return translator
